/* mfaim
 * Copyright (c) 1999 Josh Myer <josh@joshisanerd.com>
 * console-mode AIM client
 *
 * mfaim is based on naim;
 *  naim is Copyright (c) 1998, 1999 Daniel Reed <n@ml.org>
 * 
 * mfaim also uess libfaim,
 *  FAIM is Copyright 1998 Adam Fritzler <afritz@iname.com>
 *  FAIM uses code Copyright 1998 Brock Wilcox <awwaiid@iname.com>
 */

#include <mfaim.h>

extern WINDOW	*win_status;
extern buddywin_t	nullbwin, *curbwin;
extern faimconf_t	faimconf;
extern mfaimses_t       mfaimses;
struct aim_conn_t	*nfaim_conn;
extern struct aim_session_t mfaim_session;

char *mfaim_genverstring(void)
{
  static char str[151];

  snprintf(str, 150, 
	   "Client information: mfaim-pre2.4-cvs1 [Probable build time: %s %s] / libfaim %s",
	   __DATE__, __TIME__, aim_getbuildstring());

  return str;
}

void mfaim_budprof(struct aim_conn_t *conn) {
  char	*buddies = NULL;
  
  printf("Uploading buddy list and profile...\n");
  buddies = rmakebuddyar(&faimconf);
  if(buddies) {
    aim_bos_setbuddylist(&mfaim_session, conn, buddies);
    free(buddies);
  }
  if (faimconf.profile == NULL) {
    char *verstring;
    verstring = mfaim_genverstring(); /* returns pointer to static buffer! */
    faimconf.profile = (char *)calloc(1,strlen(verstring)+1);
    strncpy(faimconf.profile, verstring, strlen(verstring));
  }
  aim_bos_setprofile(&mfaim_session, conn, faimconf.profile, NULL, AIM_CAPS_CHAT | AIM_CAPS_VOICE | AIM_CAPS_GETFILE | AIM_CAPS_SENDFILE | AIM_CAPS_IMIMAGE);
  aim_conn_setlatency(conn, 1); 
}


nFAIM_HANDLER(mfaim_handle_connerr) {
  va_list ap;
  unsigned short code;
  char *msg = NULL;

  ap = va_start(ap, command);
  code = va_arg(ap, unsigned short);
  msg = va_arg(ap, char *);
  va_end(ap);

  /* we'll let josh take it from here -- mid */
  printf("BOS connection lost: %s (Code 0x%04x)\n", msg, code);
  aim_conn_kill(sess, &command->conn);

  return 1;
}

nFAIM_HANDLER(mfaim_redirect_handle) {
  va_list	ap;
  int	serviceid;
  char	*ip, *cookie;
  
  va_start(ap, command);
  serviceid = va_arg(ap, int);
  ip = va_arg(ap, char *);
  cookie = va_arg(ap, char *);
   
  switch(serviceid) {
  case 0x0005: /* Advertisements */
    mfaim_budprof(command->conn);
    aim_seticbmparam(sess, command->conn);
    aim_bos_clientready(sess, command->conn);
    
    /*
     * Yes, we're going to start up chatnav.  But all we're
     * doing is requesting chat rights, nothing more.  The
     * connection should get closed up as soon as thats over.
     */
    aim_bos_reqservice(sess, command->conn, AIM_CONN_TYPE_CHATNAV);

    printf("Client ready.\n");
    nfaim_conn = command->conn;
    mfaimses.connected = 1;
    mfaimses.online = time(NULL);
    break;
#if 0
  case 0x0007: { /* Auth */
    struct aim_conn_t	*tstconn;
    
    tstconn = aim_newconn(sess, AIM_CONN_TYPE_AUTH, ip);
    if ((tstconn == NULL) ||
 	(tstconn->status >= AIM_CONN_STATUS_RESOLVERR))
      debug_printf("Unable to reconnect to authorizer\n");
    else
      aim_auth_sendcookie(sess, tstconn, cookie);
    break;
  }
#endif
  case 0x000d: { /* Chat Nav */
    struct aim_conn_t *tstconn = NULL;
    tstconn = aim_newconn(sess, AIM_CONN_TYPE_CHATNAV, ip);
    if ( (tstconn==NULL) || (tstconn->status >= AIM_CONN_STATUS_RESOLVERR))
      {
	fprintf(stderr, "faimtest: unable to connect to chatnav server\n");
	if (tstconn) aim_conn_kill(sess, &tstconn);
	return 1;
      }
#if 0
    aim_conn_addhandler(sess, tstconn, AIM_CB_FAM_CTN, AIM_CB_SPECIAL_DEFAULT, aim_parse_unknown, 0);
    aim_conn_addhandler(sess, tstconn, AIM_CB_FAM_GEN, AIM_CB_SPECIAL_DEFAULT, aim_parse_unknown, 0);
#endif
    aim_conn_addhandler(sess, tstconn, 0x0001, 0x0003, mfaim_serverready_handle, 0);
    aim_auth_sendcookie(sess, tstconn, cookie);
    break;
  }
  case 0x000e: 
    {
      char *roomname = NULL;
      struct aim_conn_t *tstconn = NULL;
      
      roomname = va_arg(ap, char *);
      
      tstconn = aim_newconn(sess, AIM_CONN_TYPE_CHAT, ip);
      if ( (tstconn==NULL) || (tstconn->status >= AIM_CONN_STATUS_RESOLVERR))
	{
	  fprintf(stderr, "faimtest: unable to connect to chat server\n");
	  if (tstconn) aim_conn_kill(sess, &tstconn);
	  return 1;
	}		
      printf("Connected to the Chat service for room %s\n", roomname);
      
      /*
       * We must do this to attach the stored name to the connection!
       */
      aim_chat_attachname(tstconn, roomname);
      
      aim_conn_addhandler(sess, tstconn, 0x0001, 0x0003, mfaim_chatsvrready, 0);
      aim_auth_sendcookie(sess, tstconn, cookie);
      break;
    }	
  default:
    printf("Unknown redirect %#04X\n", serviceid);
    break;
  }

  va_end(ap);
  return(1);
}

nFAIM_HANDLER(mfaim_serverready_handle) {
  switch (command->conn->type) {
  case AIM_CONN_TYPE_BOS:
    aim_bos_reqrate(sess, command->conn);
    aim_bos_ackrateresp(sess, command->conn);
    aim_bos_setprivacyflags(sess, command->conn, 0x00000003);
    aim_bos_reqservice(sess, command->conn, AIM_CONN_TYPE_ADS);
    aim_bos_setgroupperm(sess, command->conn, 0x1f);
    break;
  case AIM_CONN_TYPE_CHATNAV:
    aim_conn_addhandler(sess, command->conn, AIM_CB_FAM_CTN, AIM_CB_CTN_INFO, mfaim_chatnav_info, 0);
    aim_bos_reqrate(sess, command->conn);
    aim_bos_ackrateresp(sess, command->conn);
    aim_chatnav_clientready(sess, command->conn);
    aim_chatnav_reqrights(sess, command->conn); /* once we get this, killconn*/
    break;
  default:
    printf("Unknown connection type on serverready\n");
    break;
  }
  return(1);
}

nFAIM_HANDLER(mfaim_connectack) {
  aim_auth_sendcookie(sess, command->conn, sess->logininfo.cookie);
  return 1;	
}

nFAIM_HANDLER(mfaim_motd) {
  char *msg;
  u_short id;
  va_list ap;
  
  va_start(ap, command);
  id = va_arg(ap, u_short);
  msg = va_arg(ap, char *);
  va_end(ap);
  
  printf("AIM Message of the Day: %s\n", msg);

  return 1;
}

nFAIM_HANDLER(mfaim_auth_success) {
  struct aim_conn_t	*bosconn = NULL;

  if (sess->logininfo.errorcode) {
    wshutitdown();
    
    printf("Screen name: %s\n", sess->logininfo.screen_name);
    printf("Error URL: %s\n", sess->logininfo.errorurl);
    printf("Error code: 0x%02x (", sess->logininfo.errorcode);

    switch(sess->logininfo.errorcode)
      {
      case 0x0001: printf("Invalid Screen Name"); break;
      case 0x0005: printf("Invalid Password"); break;
      case 0x0018: printf("Login Rate Reservation"); break;
      }
    printf(")\n");
    aim_conn_kill(sess, &command->conn);
    exit(0); /* there's still memory to be freed, but what do we care... */
  }

  printf("Auth successful, logging in to %s:\n", sess->logininfo.BOSIP);
  printf("          Screen name: %s\n", sess->logininfo.screen_name);
  if (sess->logininfo.email)
    printf("       Email addresss: %s\n", sess->logininfo.email);
  if (sess->logininfo.regstatus)
    printf("  Registration status: %02i\n", sess->logininfo.regstatus);
  debug_printf("Connecting to %s, closing auth connection.\n",
	       sess->logininfo.BOSIP);

  aim_conn_kill(sess, &command->conn);

  if ((bosconn = aim_newconn(sess, AIM_CONN_TYPE_BOS, sess->logininfo.BOSIP))
      == NULL) {
    printf("Could not connect to BOS: internal error\n");
    return 1;
  } else if (bosconn->status != 0) {
    printf("Could not connect to BOS\n");
    aim_conn_kill(sess, &bosconn);
    return 1;
  } else {
    aim_conn_addhandler(sess, bosconn, 0, 1, mfaim_connectack, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_SERVERREADY, mfaim_serverready_handle, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_RATEINFO, NULL, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_REDIRECT, mfaim_redirect_handle, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_MOTD, mfaim_motd, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_STS, AIM_CB_STS_SETREPORTINTERVAL, NULL, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_BUD, AIM_CB_BUD_ONCOMING, mfaim_buddy_coming, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_BUD, AIM_CB_BUD_OFFGOING, mfaim_buddy_going, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_MSG, AIM_CB_MSG_INCOMING, mfaim_im_handle, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_LOC, AIM_CB_LOC_ERROR, mfaim_handler_loc_error, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_MSG, AIM_CB_MSG_MISSEDCALL, mfaim_handler_msg_missed, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_RATECHANGE, mfaim_handler_ratechange, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_MSG, AIM_CB_MSG_ERROR, mfaim_handler_msg_error, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_LOC, AIM_CB_LOC_USERINFO, mfaim_userinfo_handler, 0);
    aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_CONNERR, mfaim_handle_connerr, 0);

    debug_printf("BOS connection established\n");
 
    return(1);
  }
}

nFAIM_HANDLER(mfaim_buddy_coming) {
  struct aim_userinfo_s *userinfo;
  buddywin_t *bwin = curbwin;
  va_list	ap;

  va_start(ap, command);
  userinfo = va_arg(ap, struct aim_userinfo_s *);
  va_end(ap);

  bcoming(userinfo->sn);
  /* XXX: how do i get the away message?   (you dont --mid) */
  bwin = bgetwin(userinfo->sn); 

  if((userinfo->class & AIM_CLASS_AWAY) == AIM_CLASS_AWAY) {
    if((bwin->away) == 0) {
      bwin->away = 1;
      mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "%s is away<BR>", userinfo->sn);
    }
    // wdirty(bwin);
  } else {
    if((bwin->away) == 1) {
      bwin->away = 0;
      mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "%s is no longer away<BR>", userinfo->sn);
    }
  }


  return(1);
}

nFAIM_HANDLER(mfaim_buddy_going) {
  va_list ap;
  char *sn;
  
  va_start(ap, command);
  sn = va_arg(ap, char *);
  va_end(ap);

  bgoing(sn);

  return(1);
}

nFAIM_HANDLER(mfaim_im_handle) {
  buddywin_t	*bwin;
  time_t	t = 0;
  struct tm	*tmptr = NULL;
  ignorelist_t	*idiot = NULL;
  char	*msg = NULL;
  u_int icbmflags = 0;
  struct aim_userinfo_s *userinfo;
  va_list	ap;
  int channel;
  u_short encflag1,encflag2;

  t = time(NULL);
  assert(t != 0);
  tmptr = localtime(&t);
  assert(tmptr != NULL);
  idiot = faimconf.idiots;

  va_start(ap, command);
  channel = va_arg(ap, int);

  /* XXX: integrate this with all the logging and crap...josh can do that. */
  /*      josh wasn't paying attention.  we'll mess with it some more and see if */
  /*      he notices this time. */
  if (channel == 0x02) {
    int reqclass = 0;
    
    reqclass = va_arg(ap, unsigned short);
    
    switch (reqclass) {
    case AIM_RENDEZVOUS_VOICE: {
      userinfo = va_arg(ap, struct aim_userinfo_s *);
      va_end(ap);
      
      bwin = wgetwin(userinfo->sn);

      wdirty(bwin);
      mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "Voice Chat Invitation from <B>%s</B><br>",userinfo->sn);

      logim("VOICE INV", bwin->winname, msg);
      
      wbuddy_update();
      return(1);
      break;
    }
    case AIM_RENDEZVOUS_FILETRANSFER: {
      printf("file transfer request!\n");
      return 1;
      break;
    }
    case AIM_RENDEZVOUS_CHAT_EX3:
    case AIM_RENDEZVOUS_CHAT_EX4:
    case AIM_RENDEZVOUS_CHAT_EX5: {
      char *msg,*encoding,*lang;
      struct aim_chat_roominfo *roominfo;
      
      userinfo = va_arg(ap, struct aim_userinfo_s *);
      roominfo = va_arg(ap, struct aim_chat_roominfo *);
      msg = va_arg(ap, char *);
      encoding = va_arg(ap, char *);
      lang = va_arg(ap, char *);
      va_end(ap);

      bwin = wgetwin(userinfo->sn);
      
      wdirty(bwin);

      mfaim_dprintf(bwin->win, IMWIN_TEXT_R,"Chat Invitation from <B>%s</B>: %s (Room: %s, Exchange %x, Instance: %x)<br>",userinfo->sn,msg,roominfo->name, roominfo->exchange, roominfo->instance);

      logim("INVITE", roominfo->name, bwin->winname);

      wbuddy_update();  	  

      return(1);

      break;
    }
    default: {
      printf("Unsupported reqclass 0x%04x!\n", reqclass);
      break;
    }
    } /* switch */
    
  }
  
  userinfo = va_arg(ap, struct aim_userinfo_s *);
  msg = va_arg(ap, char *);
  icbmflags = va_arg(ap, u_int);
  encflag1 = va_arg(ap, u_short);
  encflag2 = va_arg(ap, u_short);
  va_end(ap);
    
  logim(userinfo->sn, sess->logininfo.screen_name, msg);
  /* XXX: libfaim ignores */
  /* I never much liked the current ignore code anyway... */
  /* Adam, we need to sit down and talk about passing structs to libfaim ;^) */
  /* yes, again. i still say i'm right. so there. */

  /* I love abusing comments.*/

  /* oh, piss off.  complain about something real :) -mid */

  while (idiot != NULL) {
    if (aimcmp(userinfo->sn, idiot->screenname) == 0) {
      if ((t - idiot->informed) > 1200) {
	char igbuf[1200];
	
	snprintf(igbuf, sizeof(igbuf), "You are"
		 " on my ignore list: <font"
		 " size=+1 absz=14>%s"
		 "</font><br>\n"
		 "Your IM was not received by me,"
		 " it was ignored. You will receive"
		 " no more IM's from me (at least"
		 " for 20 minutes).<br>\n<font"
		 " size=-1 absz=10>This IM"
		 " is automatic.</font>",
		 idiot->notes);
	mfaim_send_im_away(userinfo->sn, igbuf);
	idiot->informed = t;
      }
      return(1);
    }
    idiot = idiot->next;
  }

  if (secs_getvar_int("beeponim") == 1)
    beep();

  if (icbmflags & AIM_IMFLAGS_ACK)
    mfaim_dprintf(win_status, IMWIN_TEXT_R,"Ignoring ICBM Ack Request from <B>%s</B><br>",userinfo->sn);

  bwin = wgetwin(userinfo->sn);
  wdirty(bwin);


  if (!(icbmflags & AIM_IMFLAGS_AWAY)) {
    if (atoi(secs_getvar("timestamps")) & 0x01)
      mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "<B>%s</B>: %s<br>",userinfo->sn, msg);
    else 
      hwprintf(bwin->win, IMWIN_TEXT_R, "<B>%s</B>: %s<br>", userinfo->sn, msg);

    if (nullbwin.away && (*secs_getvar("awaymsg") != 0) && ((t - bwin->informed) > 600) && (aimcmp(userinfo->sn, sess->logininfo.screen_name) != 0)) {
      sendaway(userinfo->sn);
      bwin->informed = t;
    }
  } else {
    mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "<B>%s</B> is away: %s<br>",userinfo->sn, msg);
  }
  wbuddy_update();
  return(1);
}

nFAIM_HANDLER(mfaim_handler_loc_error) {
  printf("Requested user information for a non-existant or offline user\n");
  return (1);
};

nFAIM_HANDLER(mfaim_handler_msg_missed) {

  buddywin_t *tempbwin;
  struct tm *tmptr = NULL;
  time_t t = time(NULL);

  tmptr = localtime(&t);

  if (strlen(&(command->data[13])) > MAXSNLEN)
    return 1;  /* messed up frame */

  tempbwin = wgetwin(&(command->data[13]));
  
  mfaim_dprintf(tempbwin->win, STAT_TEXT,"Server missed some messages from <B>%s</B> because they were sent too fast.<br>",&(command->data[13]));	

  wdirty(tempbwin);

    logim("RX ERROR", tempbwin->winname, "Server missed some messages from buddy because they were sent too fast (not RC).");
  return (1);
};

nFAIM_HANDLER(mfaim_handler_ratechange) {
  buddywin_t *tempbwin;
  struct tm *tmptr = NULL;
  time_t t = time(NULL);

  tmptr = localtime(&t);

  tempbwin = wgetwin(&(command->data[13]));

  mfaim_dprintf(tempbwin->win, IMWIN_TEXT_R,"Server missed some messages from <B>%s</B> because they were sent too fast. (Rate change)<br>",&(command->data[13]));
  logim("RX ERROR", tempbwin->winname, "Server missed some messages from buddy because they were sent too fast (RC).");
  return (1);
};

nFAIM_HANDLER(mfaim_handler_msg_error) {
  va_list ap;
  char *destsn;
  buddywin_t *tempbwin;
  struct tm *tmptr = NULL;
  time_t t = time(NULL);

  va_start(ap, command);
  destsn = va_arg(ap, char *);
  va_end(ap);

  tempbwin = wgetwin(destsn);

  tmptr = localtime(&t);

  mfaim_dprintf(tempbwin->win, STAT_TEXT,"Error sending last message to <B>%s</B>.  The destination may not be online.<br>",destsn);
  logim("SEND ERROR", tempbwin->winname, "Error sending last message. Dest offline?");

  return (1);
};

nFAIM_HANDLER(mfaim_userinfo_handler) {
  time_t	t;
  struct tm	*tmptr;
  struct aim_userinfo_s *userinfo;
  char *text_encoding, *text;
  va_list	ap;
  unsigned short infotype;

  va_start(ap, command);
  userinfo = va_arg(ap, struct aim_userinfo_s *);
  text_encoding = va_arg(ap, char *);
  text = va_arg(ap, char *);
  infotype = va_arg(ap, unsigned short);
  va_end(ap);

  if (infotype == AIM_GETINFO_GENERALINFO) {
    printf("Information about %s:\n", userinfo->sn);
    printf("  Warning level: %i%%\n", userinfo->warnlevel/10);
    printf("          Class:%s%s%s%s%s%s%s%s\n", 
	   (userinfo->class&AIM_CLASS_FREE)?" Free":"",
	   (userinfo->class&AIM_CLASS_TRIAL)?" Trial":"", 
	   (userinfo->class&AIM_CLASS_AOL)?" AOLamer":"",
	   (userinfo->class&AIM_CLASS_AWAY)?" Away":"",
	   (userinfo->class&AIM_CLASS_UNKNOWN2)?" UnknownBit2":"", 
	   (userinfo->class&AIM_CLASS_UNKNOWN4)?" UnknwonBit4":"",
	   (userinfo->class&AIM_CLASS_UNKNOWN40)?" UnknownBit40":"",
	   (userinfo->class&AIM_CLASS_UNKNOWN80)?" UnknownBit80":"");
    if (userinfo->capabilities)
      printf("Client Capabilities: %x\n", userinfo->capabilities);
    if ((t = userinfo->membersince) != 0) {
      tmptr = localtime(&t);
      assert(tmptr != NULL);
      printf("   Member since: %02i-%02i-%04i at"
	     " %02i:%02i\n", 1+tmptr->tm_mon,
	     tmptr->tm_mday, 1900+tmptr->tm_year,
	     tmptr->tm_hour, tmptr->tm_min);
    }
    if ((t = userinfo->onlinesince) != 0) {
      tmptr = localtime(&t);
      assert(tmptr != NULL);
      printf("   Online since: %02i-%02i-%04i at"
	     " %02i:%02i\n", 1+tmptr->tm_mon,
	     tmptr->tm_mday, 1900+tmptr->tm_year,
	     tmptr->tm_hour, tmptr->tm_min);
    }
    if (userinfo->idletime > 0)
      printf("      Idle time: %i minutes\n", userinfo->idletime);
    if (userinfo->sessionlen!=-1)
      printf("   Time elapsed: %ld seconds\n", userinfo->sessionlen);
    
    if ((text != NULL) && (text_encoding != NULL)) {
      printf("Profile (%s):\n", text_encoding);
      hwprintf(win_status, STAT_TEXT, "%s<br><hr>", text);
    }
    
    /* If the client is new enough to have this set, then its new
     * enough for us to request their away message as well. 
     */
    if (userinfo->class & AIM_CLASS_AWAY) 
      aim_getinfo(sess, command->conn, userinfo->sn, AIM_GETINFO_AWAYMESSAGE);

  } else if (infotype == AIM_GETINFO_AWAYMESSAGE) {
    printf("Away message for %s (%s):\n", userinfo->sn, text_encoding);
    hwprintf(win_status, STAT_TEXT, "%s<br><hr>", text);
  }

  return(1);
}



nFAIM_HANDLER(mfaim_chat_join)
{
  va_list ap;
  struct aim_userinfo_s *userinfo;
  int count = 0, i = 0;
  buddywin_t *bwin;
  time_t	t = 0;
  struct tm	*tmptr = NULL;

  bwin = wgetwin(aim_chat_getname(command->conn));

  t = time(NULL);
  tmptr = localtime(&t);

  va_start(ap, command);
  count = va_arg(ap, int);
  userinfo = va_arg(ap, struct aim_userinfo_s *);
  va_end(ap);

  wdirty(bwin);

  if(secs_getvar_int("beeponsignon"))
    beep();

  while (i < count) {
    mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "<B>%s</B> has entered %s<br>",userinfo[i++].sn, aim_chat_getname(command->conn));
    logim(aim_chat_getname(command->conn), "JOIN", userinfo[i].sn);
  }
  wbuddy_update();
  return 1;
}

nFAIM_HANDLER(mfaim_chat_leave)
{
  va_list ap;
  struct aim_userinfo_s *userinfo;
  int count = 0, i = 0;
  buddywin_t *bwin;
  time_t	t = 0;
  struct tm	*tmptr = NULL;

  bwin = wgetwin(aim_chat_getname(command->conn));

  t = time(NULL);
  tmptr = localtime(&t);

  va_start(ap, command);
  count = va_arg(ap, int);
  userinfo = va_arg(ap, struct aim_userinfo_s *);
  va_end(ap);

  wdirty(bwin);

  if(secs_getvar_int("beeponsignon"))
    beep();

  while (i < count) {
    mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "<B>%s</B> has left %s<br>",userinfo[i++].sn, aim_chat_getname(command->conn));
    logim(aim_chat_getname(command->conn), "PART", userinfo[i].sn);
  }

  wbuddy_update();
  return 1;
}


nFAIM_HANDLER(mfaim_chat_infoupdate)
{
  va_list ap;
  struct aim_userinfo_s *userinfo;
  struct aim_chat_roominfo *roominfo;
  char *roomname;
  int usercount,i;
  char *roomdesc;
  buddywin_t *bwin;
  time_t	t = 0;
  struct tm	*tmptr = NULL;
  unsigned short unknown_c9, unknown_d2, unknown_d5, maxmsglen;
  unsigned long creationtime;


  bwin = wgetwin(aim_chat_getname(command->conn));

  t = time(NULL);
  tmptr = localtime(&t);

  va_start(ap, command);
  roominfo = va_arg(ap, struct aim_chat_roominfo *);
  roomname = va_arg(ap, char *);
  usercount= va_arg(ap, int);
  userinfo = va_arg(ap, struct aim_userinfo_s *);
  roomdesc = va_arg(ap, char *);
  unknown_c9 = va_arg(ap, unsigned short);
  creationtime = va_arg(ap, unsigned long);
  maxmsglen = va_arg(ap, unsigned short);
  unknown_d2 = va_arg(ap, unsigned short);
  unknown_d5 = va_arg(ap, unsigned short);
  va_end(ap);

  mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "Room information for <B>%s</B> (exchange %x/%x):<br>",roominfo->name, 	roominfo->exchange, roominfo->instance);

  mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "        Desc: %s<br>",roomdesc);

  mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "        Occupants: %d<br>",usercount);
  
#if 0  /* why are we not doing this? -mid */
  i = 0; 
  while (i < usercount)
    mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "          %s<br>",userinfo[i++].sn);
#endif

  mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "        Unknown_c9: 0x%04x<br>", unknown_c9);
  
  t = creationtime;
  tmptr = localtime(&t);
  if (tmptr) {
    mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "        Creation time: %02d-%02d-%04d %02d:%02d<br>", 
		  1+tmptr->tm_mon, tmptr->tm_mday, 
		  1900+tmptr->tm_year, tmptr->tm_hour, 
		  tmptr->tm_min);
  }
  mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "        Max message length: %d bytes<br>", maxmsglen);
  mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "        Unknown_d2: 0x%04x<br>", unknown_d2);
  mfaim_dprintf(bwin->win, IMWIN_TEXT_R, "        Unknown_d5: 0x%02x<br>", unknown_d5);

  wbuddy_update();
  return 1;
}

nFAIM_HANDLER(mfaim_chat_incomingmsg)
{
  va_list ap;
  struct aim_userinfo_s *userinfo;
  char *msg;
  buddywin_t	*bwin;
  time_t	t = 0;
  struct tm	*tmptr = NULL;
  int pos = 0;

  t = time(NULL);
  tmptr = localtime(&t);

  va_start(ap, command);
  userinfo = va_arg(ap, struct aim_userinfo_s *);	
  msg = va_arg(ap, char *);
  va_end(ap);

  bwin = wgetwin(aim_chat_getname(command->conn));

    wdirty(bwin);

    pos = 0;

    while(*(msg+pos) == '<')
      pos += parsehtml(NULL,msg+pos,NULL,NULL,NULL)+1;

    if(strstr(msg+pos, "/me ") == msg+pos) {
      if (atoi(secs_getvar("timestamps")) & 0x04)
	mfaim_dprintf(bwin->win, (aimcmp(faimconf.screenname,userinfo->sn)?IMWIN_TEXT_R:IMWIN_TEXT_U), "*** <B>%s</B> %s<br>",userinfo->sn,msg+4+pos);
      else
	hwprintf(bwin->win, (aimcmp(faimconf.screenname,userinfo->sn)?IMWIN_TEXT_R:IMWIN_TEXT_U), "*** <B>%s</B> %s<br>",userinfo->sn,msg+4+pos);
    } else {
      if (atoi(secs_getvar("timestamps")) & 0x04)
	mfaim_dprintf(bwin->win, (aimcmp(faimconf.screenname,userinfo->sn)?IMWIN_TEXT_R:IMWIN_TEXT_U), "<B>&lt;%s&gt;</B> %s<br>",userinfo->sn,msg);      
      else
	hwprintf(bwin->win, (aimcmp(faimconf.screenname,userinfo->sn)?IMWIN_TEXT_R:IMWIN_TEXT_U), "<B>&lt;%s&gt;</B> %s<br>",userinfo->sn,msg);  
    }

    logim(userinfo->sn, aim_chat_getname(command->conn), msg);

    if((aimcmp(faimconf.screenname,userinfo->sn) != 0) && (secs_getvar_int("beeponim") == 1))
      beep();

  wbuddy_update();
  return 1;
}

void openchatwin(char *roomname) {
  buddywin_t *bwin = NULL;
    
  bnewwin(roomname);
  bwin = wgetwin(roomname);
  bwin->offline = 0;
  bwin->notbuddy = 0;
  bwin->chat = 1;
 
  wbuddy_update();
  return;
}

nFAIM_HANDLER(mfaim_chatsvrready)
{
  aim_conn_addhandler(sess, command->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_USERJOIN, mfaim_chat_join, 0);
  aim_conn_addhandler(sess, command->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_USERLEAVE, mfaim_chat_leave, 0);
  aim_conn_addhandler(sess, command->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_ROOMINFOUPDATE, mfaim_chat_infoupdate, 0);
  aim_conn_addhandler(sess, command->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_INCOMINGMSG, mfaim_chat_incomingmsg, 0);
  aim_bos_reqrate(sess, command->conn);
  aim_bos_ackrateresp(sess, command->conn);
  aim_chat_clientready(sess, command->conn);
  
  /*
   * Is this what we want to do ?
   */
  openchatwin(aim_chat_getname(command->conn));
  return 1;
}

int mfaim_chatnav_info(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
  u_short type;
  va_list ap;
  time_t	t = 0;
  struct tm	*tmptr = NULL;

  t = time(NULL);
  tmptr = localtime(&t);

  ap = va_start(ap, command);
  type = va_arg(ap, u_short);

  switch(type)
    {
    case 0x0002:
      {
	int maxrooms;
	struct aim_chat_exchangeinfo *exchanges;
	int exchangecount,i = 0;
	
	maxrooms = va_arg(ap, u_char);
	exchangecount = va_arg(ap, int);
	exchanges = va_arg(ap, struct aim_chat_exchangeinfo *);
	va_end(ap);

	mfaim_dprintf(win_status, STAT_TEXT, "Maximum concurrent chat rooms: %d<br>",maxrooms);

	mfaim_dprintf(win_status, STAT_TEXT, "Chat room exchanges: (%d total)<br>",exchangecount);

	

	while (i < exchangecount)
	  {
	    mfaim_dprintf(win_status, STAT_TEXT, "        %d: %s (%s/%s)<br>",exchanges[i].number,	exchanges[i].name, exchanges[i].charset1, exchanges[i].lang1);
	    i++;
	  }

	/*
	 * It should be okay to close this now.
	 */
	rjoin();
	aim_conn_kill(sess, &command->conn);
      }
      break;
    default:
      va_end(ap);
      printf("faimtest: chatnav info: unknown type (%04x)\n", type);
    }
  wbuddy_update();
  return 1;
}
