
/* mfaim
 * Copyright (c) 1999 Josh Myer <josh@joshisanerd.com>
 * console-mode AIM client
 *
 * mfaim is based on naim;
 *  naim is Copyright (c) 1998, 1999 Daniel Reed <n@ml.org>
 * 
 * mfaim also uess libfaim,
 *  FAIM is Copyright 1998 Adam Fritzler <afritz@iname.com>
 *  FAIM uses code Copyright 1998 Brock Wilcox <awwaiid@iname.com>
 */

#include <mfaim.h>

extern buddywin_t	*curbwin;
extern faimconf_t	faimconf;

#define CHECKTAG(tag)	(strcasecmp(tagbase, (tag)) == 0)
#define CHECKAMP(tag)	(strcasecmp(tagbuf, (tag)) == 0)

unsigned long parsehtml_tag(WINDOW *window, char *text, char *inbold, 
			    char *initalic, char *inunderline) {
  
  char	tagbuf[20], argbuf[1024], *textsave = text, *tagbase, inquote = 0;
  int	tagpos = 0;
  unsigned long retval = 1;

  bzero(tagbuf, sizeof(tagbuf));
  bzero(argbuf, sizeof(argbuf));
  
  while (isspace(*text))
    text++;
  while ((*text) && (!isspace(*text)) && (*text != '>')) {

    if(*text == '"') /* handle quoted >s.. html is odd. */
      while(*text != '"')
	text++;
    
    if (tagpos < sizeof(tagbuf))
      tagbuf[tagpos++] = *text;
    text++;
  }

  while (isspace(*text))
    text++;
  tagpos = 0;

  while ((*text) && (inquote || (*text != '>' && !inquote))) {

    if(*text == '"') /* handle quoted >s.. html is odd. */
      inquote = (inquote == 0);

    if (tagpos < sizeof(argbuf))
      argbuf[tagpos++] = *text;
    text++;
  }
  tagbase = tagbuf;
  if (tagbuf[0] == '/')
    tagbase++;

  if(window != NULL) {
    //    wattrset(window, A_BOLD | COLOR_PAIR(IMWIN_STATUS));
    if (CHECKTAG("B")) {
      if(inbold != NULL)
	*inbold = (*inbold == 0);
    } else if CHECKTAG("I") {
      if(inbold != NULL)
	*initalic = (*initalic == 0);
    } else if CHECKTAG("U") {
      if(inunderline != NULL)
	*inunderline = (*inunderline == 0);
    } else if CHECKTAG("A") {
      if (tagbuf[0] == '/')
	waddstr(window, "]");
      else {
	waddstr(window, "[LINK:");
	waddstr(window, argbuf);
	waddstr(window, ":");
      }
    } else if CHECKTAG("IMG") {
      waddstr(window, "[IMAGE:");
      waddstr(window, argbuf);
      waddstr(window, "]");
    } else if CHECKTAG("BR") {
      waddstr(window, "\n");
    } else if CHECKTAG("HR") {
      waddstr(window, "----------------\n");
    } else if CHECKTAG("FONT") {
      ;
    } else if CHECKTAG("PRE") {
      ;
    } else if CHECKTAG("P") {
      ;
    } else if CHECKTAG("HTML") {
      ;
    } else if CHECKTAG("BODY") {
      ;
    } else
      retval = 0;
  }
  if (retval != 0)
    retval = (unsigned long)(text-textsave+1);
  return(retval);
}

unsigned long	parsehtml_amp(WINDOW *window, char *text) {
  char	tagbuf[20], *textsave = text;
  int	tagpos = 0;
  unsigned long retval = 1;

  bzero(tagbuf, sizeof(tagbuf));
  
  while (isspace(*text))
    text++;
  while ((*text) && (!isspace(*text)) && (*text != ';') && (*text != '\n')) {
    if (tagpos < sizeof(tagbuf))
      tagbuf[tagpos++] = *text;
    else
      return(0);
    text++;
  }
  if (*text != ';')
    return(0);
  if(window != NULL) {
    //    wattrset(window, A_BOLD | COLOR_PAIR(IMWIN_STATUS));
    if CHECKAMP("NBSP") {
      waddstr(window, " ");
    } else if CHECKAMP("AMP") {
      waddstr(window, "&");
    } else if CHECKAMP("LT") {
      waddstr(window, "<");
    } else if CHECKAMP("GT") {
      waddstr(window, ">");
    } else
      retval = 0;
  }
  if(retval != 0)
    retval = (unsigned long)(text - textsave +1);
  return(retval);
}

unsigned long	parsehtml(WINDOW *w, char *a, char *b, char *c, char *d) {
  if (*a == '<')
    return(parsehtml_tag(w, a+1, b, c, d));
  else if (*a == '&')
    return(parsehtml_amp(w, a+1));
  else
    return(0);
}

int hwprintf_(WINDOW *window, int pair, char *buf);
int mfaim_dprintf(WINDOW *window, int pair, const char *format, ...) {
  char *buf, *outbuf;
  va_list msg;
  time_t t = 0;
  struct tm *tmptr = NULL;
  int retval = 0, size = 2048;

  t = time(NULL);
  tmptr = localtime(&t);

  buf = (char *)calloc(1,strlen(format)+25); /* 20 is arbitrarily large */
  snprintf(buf, (strlen(format)+25), "[%02i:%02i:%02i] %s", tmptr->tm_hour, tmptr->tm_min,tmptr->tm_sec,format);

  outbuf = (char *)malloc(size);
  bzero(outbuf, size);

  va_start(msg, format);

  while(vsnprintf(outbuf,size,buf, msg) == -1) {
    size += 256;
    realloc(outbuf, size);
    bzero(outbuf, size);
  }
 
  retval = hwprintf_(window, pair, outbuf);
 
  free(outbuf);
  free(buf);
  return(retval);
}

int hwprintf(WINDOW *window, int pair, const char *format, ...) {
  char *outbuf;
  va_list msg;
  int retval = 0, size = 1024;

  outbuf = (char *)malloc(size);
  bzero(outbuf, size);

  va_start(msg, format);

  while(vsnprintf(outbuf,size,format, msg) == -1) {
    size += 256;
    realloc(outbuf, size);
    bzero(outbuf, size);
  }
 
  retval = hwprintf_(window, pair, outbuf);
 
  free(outbuf);
  return(retval);
}

int hwprintf_(WINDOW *window, int pair, char *buf) {
  char inbold = 0, initalic = 0, inunderline = 0;
  int len = 0, pos = 0, lines = 0, lwp = -1, sp1 =0, sp2 = 0;

  len = strlen(buf);

  while(pos < len) {
    if( (buf[pos] == '<') || (buf[pos] == '&') ) {
      if ((buf[pos+1] == 'B') || (buf[pos+1] == 'b'))
	if ((buf[pos+2] == 'R') || (buf[pos+1] == 'r'))
	  lwp = sp1 = sp2 = 0;
      if (buf[pos+1] == 'P')
	lwp = sp1 = sp2 = 0;
      pos += parsehtml(NULL, buf+pos, NULL, NULL, NULL);
    }
    if((buf[pos] == ' ') || (buf[pos] == '\n') || (buf[pos] == 0)) {
      sp1 = sp2;
      sp2 = lwp;

      if(buf[pos+1] == 0) { /* optimize for a funny but annoying case */
       	buf[pos] = 0;
	break;
      }

      if ((sp1 % WSTATUS_X) > (sp2 % WSTATUS_X)) {
	buf[pos - (sp2-sp1)] = '\n';
	lwp = sp2-sp1+1;
	sp1 = sp2 = 0;
      }
    }  
    lwp++;
    pos++;
  }

  pos = -1;

  while (++pos < len) {
    wattrset(window, (inbold?A_BOLD:0)
	     | (inunderline?A_UNDERLINE:0)
	     | (initalic?A_REVERSE:0) | COLOR_PAIR(pair));
    if ((buf[pos] == '<') || (buf[pos] == '&')) {
      unsigned long skiplen = 0;
      if ((skiplen = parsehtml(window, buf+pos, &inbold,&initalic, &inunderline)) == 0) {
	//	wattrset(window, A_BOLD | COLOR_PAIR(IMWIN_STATUS));
	waddch(window, buf[pos]);
	continue;
      }
      pos += skiplen;
      continue;
    } else {
      waddch(window, buf[pos]);
      continue;
    }
    wattrset(window, 0);
  }
  return(lines);
}
